/*
 *	Copyright (C) 2000 Hitachi, Ltd.
 *	All rights reserved.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: gbsplit.c,v 1.4 2001/09/06 06:23:20 nis Exp $"; */

#include <assert.h>
#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <strings.h>

char *progname;

static void usage(void);
static void gbsplit(int);
static void gbsplitA(int);

static void
usage()
{
	fprintf(stderr, "usage: %s [-1]\n", progname);
	exit(1);
}

main(argc, argv)
char *argv[];
{
	int c1 = 0;

	if (progname = rindex(argv[0], '/')) {
		progname++;
	}
	else {
		progname = argv[0];
	}

	if (argc!=1 && argc!=2) {
		usage();
	}
	if (argc==2) {
		if (!strcmp(argv[1], "-1")) {
			c1 = 1;
		}
		else {
			usage();
		}
	}

	if (!strcmp(progname, "gbsplit")) {
		gbsplit(c1);
	}
	else if (!strcmp(progname, "gbsplitA")) {
		gbsplitA(c1);
	}
	else {
		fprintf(stderr, "configuration error\n");
		exit(1);
	}
	return 0;
}

static char word[8192];

static void
gbsplit(c1)
{
	int c, i;

/*
    tr/(){}[]|="`';<>?\/\\/ /;
    tr/A-Z/a-z/;
    tr/\.//d;
    s/[,:]\s/ /g;
    split(/\s+/);
*/

	for (i=0; (c = getchar())!=EOF; ) {
		switch (c) {
		case '(':
		case ')':
		case '{':
		case '}':
		case '[':
		case ']':
		case '|':
		case '=':
		case '"':
		case '`':
		case '\'':
		case ';':
		case '<':
		case '>':
		case '?':
		case '/':
		case '\\':
			c = ' ';
		}
		c = tolower(c);
		if (isspace(c)) {
			if (i>0 && (word[i-1]==',' || word[i-1]==':')) {
				i--;
			}
			if (i>0) {
				assert(i<sizeof word);
				word[i] = '\0';
				if (c1) {
					printf("%s\n", word);
				}
				else {
					printf("%s %s\n", word, word);
				}
				fflush(stdout);
				i = 0;
			}
		}
		else if (c=='.') {
			/* do nothing */
		}
		else if (i<sizeof word-1) {
			word[i++] = c;
		}
		else {
			/* do nothing */
		}
	}
}

static void
gbsplitA(c1)
{
	int i, c;
	int m;
#define	A	1
#define	N	2

/*
	tr A-Z a-z
	[a-z]* [0-9]*
*/

	m = 0;
	for (i=0; (c = getchar())!=EOF; ) {
		int n;
		c = tolower(c);
		if (isalpha(c)) {
			n = A;
		}
		else if (isdigit(c)) {
			n = N;
		}
		else {
			n = 0;
		}
		if (m!=n) {
			if (m==A || m==N) {
				assert(i<sizeof word);
				word[i] = '\0';
				if (c1) {
					printf("%s\n", word);
				}
				else {
					printf("%s %s\n", word, word);
				}
				fflush(stdout);
				i = 0;
			}
			m = n;
		}
		if ((n==A || n==N) && i<sizeof word-1) {
			word[i++] = c;
		}
	}
}
