/*
 *	Copyright (C) 1997 Advanced Research Laboratory, Hitachi, Ltd.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: killngw.c,v 1.9 2001/09/06 06:23:21 nis Exp $"; */

#include <sys/param.h>

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>

#include <geta/getamu.h>
#include <geta/srvmu.h>
#include <geta/iso2022jp2.h>
#include <geta/jis.h>
#include <geta/gaiji.h>
#include <geta/smap.h>

#define	BSIZE	8192

char *progname = NULL;

struct _w {
	char *name;
	int df;
};

static int isstopword(char *);
static int readstopwords(char *);

static int sound(char *);

#define NSR_INVALID_ESCS	1
#define NSR_INVALID_HDHGAI	2
#define NSR_INVALID_JIS		3
#define NSR_INVALID_JPN		4
#define NSR_INVALID_ASCII	5
#define NSR_ONLY_KANA		6
#define NSR_ONLY_NGC		7

char allowc[256];

int reason;
char *reason_name[] = {
	"no error",
	"Invalid escape sequence found.",
	"Invalid HDH-gaiji character found.",
	"Invalid JIS character found.",
	"Invalid Japanese letters found.",
	"Invalid ASCII character found.",
	"Contains only kana and length is less than 2.",
	"Contains only nogood characters.",
};

static void init_allowc(void);

main(argc, argv)
char *argv[];
{
	int verbose = 0;
	static char buf[BSIZE];
	int cwindex;

	init_allowc();
	if (progname=rindex(argv[0], '/')) {
		progname++;
	}
	else {
		progname = argv[0];
	}
	if (argc>1 && !strcmp("-v", argv[1])) {
		argc--, argv++;
		verbose = 1;
	}

	if (argc!=2 && argc!=3) {
		fprintf(stderr, "Usage: %s [-v] field(0..) [stopwords]\n", progname);
		exit(1);
	}

	cwindex = atoi(argv[1]);

	if (argc==3) {
		if (readstopwords(argv[2])==-1) {
			exit(1);
		}
	}

	for (; fgets(buf, sizeof buf, stdin); ) {
		char *ss[32];	/* magic number */
		char *cw;
		int c, i;

		chop(buf);
		c = splitv(buf, " ", ss, nelems(ss));
		cw = ss[cwindex];
		if (c<=cwindex) {
			if (verbose) {
				fprintf(stderr, "warning: too few fields: %d\n", c);
			}
			goto PRINTIT;
		}
		if (*cw == CSEP) {
			goto PRINTIT;
		}
		if (isstopword(cw)) {
			if (verbose) {
				fprintf(stderr, "stopword: ");
				ppstr(stderr, cw, strlen(cw));
				fprintf(stderr, "\n");
			}
			continue;
		}
		if (!sound(cw)) {
			if (verbose) {
				fprintf(stderr, "not sound: reason: %s, str=\"", reason_name[reason]);
				ppstr(stderr, cw, strlen(cw));
				fprintf(stderr, "\"\n");
				
			}
			continue;
		}

	PRINTIT:
		for (i=0; i<c; i++) {
			printf("%s ", ss[i]);
		}
		printf("\n");
		fflush(stdout);
	}
	return 0;
}

static int
scompar(va, vb)
const void *va, *vb;
{
	char * const *a, * const *b;
	a = va;
	b = vb;
	return strcmp(*a, *b);
}

char **stopwords = NULL;
int nstopwords = 0;

static int
readstopwords(file)
char *file;
{
	static char buf[BSIZE];
	FILE *f;
	int i, n;
	if (!(f=fopen(file, "r"))) {
		perror(file);
		return -1;
	}
	for (n=0; fgets(buf, sizeof buf, f); n++) ;
	rewind(f);
	if (n && !(stopwords = calloc(n, sizeof *stopwords))) {
		perror("calloc");
		fprintf(stderr, "%s: Panic: Cannot alloc memory for stopwords(%d,%d). Stop.\n", progname, n, sizeof *stopwords);
		return -1;
	}
	for (i=0; fgets(buf, sizeof buf, f); i++) {
		chop(buf);
		if (!(stopwords[i] = strdup(buf))) {
			perror("strdup(3)");
			return -1;
		}
	}
	if (i!=n) {
		fprintf(stderr, "Internal Error: Stop\n");
		return -1;
	}
	nstopwords = n;
	qsort(stopwords, nstopwords, sizeof *stopwords, scompar);
	fclose(f);
	return 0;
}

static int
isstopword(w)
char *w;
{
	if (nstopwords==0) {
		return 0;
	}
	return bsearch(&w, stopwords, nstopwords, sizeof *stopwords, scompar)!=NULL;
}

static int
sound(s)
char *s;
{
	int shift, shift_prev = 0, f;
	char *p;
	int onlykana = 1;
	int onlyngc = 1;

	reason = 0;

	shift = SHIFT_SB;
	f = 'B';

	for (p = s; *p; ) {
		if (*p==033) {
			char *e;
			shift_prev = shift;
			if ((shift = pescs(p, &e))==SHIFT_ERR) {
				reason=NSR_INVALID_ESCS;
				return 0;
			}
			f = *e;
			p = e+1;
		}
		else {
			switch (shift) {
				int h, l, gh, gl, k, t;
			case SHIFT_MB:
				h = *p++;
				l = *p++;
				gh = *(p);
				gl = *(p+1);
				if (gh==GHKU + 0x20 && gl == GHTEN+0x20) {
					if (!(1+0x20<=h && h<=94+0x20 && 1+0x20<=l && l<=94+0x20)) {
						reason=NSR_INVALID_HDHGAI;
						return 0;
					}
					onlyngc = 0;
					p+=2;
					break;
				}
				if (!isvalidjis(h, l)) {
					reason=NSR_INVALID_JIS;
					return 0;
				}
				k = h - 0x20;
				t = l - 0x20;
				if (!(
					k==1 && (t==25 || t==28) ||
					k==3 && 33<=t && t<=90 || /* alphabet */
					k==4 ||	/* Hiragana */
					k==5 ||	/* Katakana */
					k==6 ||	/* Greek */
					k==7 ||	/* Cyrillic */
					16<=k	/* Kanji */
					)) {
					reason=NSR_INVALID_JPN;
					return 0;
				}
				if (k!=4 && k!=5) {
					onlykana = 0;
				}
				if (k==1 && 60<=t && t<=85 ||
				   k==2 && 26<=t && t<=89 ||
				   k==3 && 33<=t && t<=90 ||
				   4<=k && k<=7 ||
				   16<=k) {
					onlyngc = 0;
				}
				break;
			case SHIFT_SS:
				shift = shift_prev;
				/* no break */
			case SHIFT_SB:
				if (allowc[*p&0xff] || isdigit(*p)) {
					p++;
				}
				else if (isalpha(*p)) {
					p++;
					onlykana = 0;
					onlyngc = 0;
				}
				else {
					reason=NSR_INVALID_ASCII;
					return 0;
				}
				break;
			default:
				break;
			}
		}
	}
	if (p-s==2 && onlykana) {
		reason=NSR_ONLY_KANA;
		return 0;
	}
	if (onlyngc) {
		reason=NSR_ONLY_NGC;
		return 0;
	}
	return 1;
}

static void
init_allowc()
{
	bzero(allowc, sizeof allowc);
	allowc['!'] = 1;
	allowc['@'] = 1;
	allowc['#'] = 1;
	allowc['$'] = 1;
	allowc['%'] = 1;
	allowc['&'] = 1;
	allowc['-'] = 1;
	allowc['_'] = 1;
	allowc['+'] = 1;
	allowc[':'] = 1;
	allowc['\''] = 1;
	allowc[','] = 1;
	allowc['.'] = 1;
	allowc['/'] = 1;
	allowc['?'] = 1;
}
