/*
 *	Copyright (C) 1998 Advanced Research Laboratory, Hitachi, Ltd.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: mkf.c,v 1.12 2001/09/06 06:23:21 nis Exp $"; */

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <netinet/in.h>

#include <fcntl.h>
#include <netdb.h>
#include <stdio.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>

#include <geta/getamu.h>
#include <geta/srvmu.h>
#include <geta/wam.h>
#include <geta/iso2022jp2.h>
#include <geta/smap.h>
#include <geta/ifsmap.h>

char *progname;

#define	BSIZE	8192

#define	CI_U_LANGUAGE	"language"

int errno;

static void add_word(char *);
static void free_words(void);
static int compar(const void *, const void *);
static int doafile(int, char *, char *, int, char *);
static int analyse_and_add(char *, char *, int, int, char *);
static int analyse_a_phrase(char *, char *);
static void usage(void);
static void fputs_sp(char *, FILE *);

static void
usage()
{
	fprintf(stderr, "usage: %s [-n|-simulate] [-r|-raw] handle filter allkeys freqfile (-l ss_local_socket|-i ss_host:ss_port)\n", progname);
	exit(1);
}

char *fname;

FILE *mkf_out=NULL;	/* output stream */

int mkf_type = RAWXR_TYPE_COUNT;

main(argc, argv)
char *argv[];
{
	int s;
	char *localsocket = NULL;
	int type = AF_UNSPEC;
	struct hostent *h;
	char *lang;
	char *sendbuf;
	int sblen;
	int i;
	int prog;
	char *filter=NULL;
	char *handle=NULL;
	int anum;
	char key[256];	/* magic number */
	char *server_host=NULL;
	int _s_host;
	int s_port = 0;
	char *freqfile, *allkeys;
	FILE *k;

	wam_init(NULL);
	if (progname = rindex(argv[0], '/')) {
		progname++;
	}
	else {
		progname = argv[0];
	}

	for (argc--, argv++; argc>0 && *argv[0]=='-'; argc--, argv++) {
		if (!strcmp("-n", argv[0]) || !strcmp("-simulate", argv[0])) {
			mkf_type = RAWXR_TYPE_SIMULATE;
		}
		else if (!strcmp("-r", argv[0]) || !strcmp("-raw", argv[0])) {
			mkf_type = RAWXR_TYPE_RAW;
		}
	}

	if (argc!=6) {
		usage();
	}

	handle = argv[0];
	filter = argv[1];
	allkeys = argv[2];
	freqfile = argv[3];

	if (!strcmp(argv[4], "-l")) {
		localsocket = argv[5];
		type = AF_UNIX;
	}
	else if (!strcmp(argv[4], "-i")) {
		char *p;
		server_host = argv[5];
		if (!(p = rindex(server_host, ':'))) {
			usage();
		}
		*p = '\0';
		s_port = atoi(p+1);
		type = AF_INET;
	}
	else {
		usage();
	}

	if (!(lang = ci_value(handle, CI_U_LANGUAGE))) {
		fprintf(stderr, "mkf: cannot get language of %s\n", handle);
		exit(1);
	}
	if (!(lang = strdup(lang))) {
		perror("strdup");
		exit(1);
	}

	fprintf(stderr, "mkf: language = [%s]\n", lang);

	if (type==AF_UNIX) {
		if ((s = un_connect_to_server(localsocket))==-1) {
			fprintf(stderr, "cannot connect to server\n");
			exit(1);
		}
	}
	else if (type==AF_INET) {
		if (!(h = gethostbyname(server_host))) {
			perror("mkf: gethostbyname");
			exit(1);
		}
		_s_host = *(int *) *h->h_addr_list;

		if ((s = in_connect_to_server(_s_host, s_port))==-1) {
			fprintf(stderr, "cannot connect to server\n");
			exit(1);
		}
	}
	else {
		fprintf(stderr, "internal error\n");
		exit(1);
	}

	if (smap_setup(&sblen, s)==-1) {
		fprintf(stderr, "smap_setup failed\n");
		exit(1);
	}

	if (!(k = fopen(allkeys, "r"))) {
		perror(allkeys);
		return -1;
	}

	for (anum=0; fgets(key, sizeof key, k); anum++) ;
	fprintf(stderr, "Number of keys = %d\n", anum);

	startprog(anum, "mkf", stderr, 30);
	if (mkf_type==RAWXR_TYPE_SIMULATE) {
		if (!(mkf_out = fopen(freqfile, "w"))) {
			perror(freqfile);
			exit(1);
		}
	}
	else {
		if (!(mkf_out = fopen(freqfile, "w"))) {
			perror(freqfile);
			exit(1);
		}
	}

	if (!(sendbuf = malloc(sblen))) {
		perror("mkf: malloc");
		return -1;
	}

	rewind(k);

	for (i=1, prog=0; i<=anum; i++) {
		char *text;
		ssize_t len;

		fgets(key, sizeof key, k);
		chop(key);

		showprog(++prog);
		if ((len = getaux_exec(filter, handle, key, &text))==-1) {
			fprintf(stderr, "getaux_exec failed\n");
			exit(1);
		}

		fprintf(mkf_out, "%c%s\n", mkf_type, key);

		if (!(text = realloc(text, strlen(text)*2+1))) {
			fprintf(stderr, "mkf: realloc failed %d/%d\n", i, anum);
			exit(1);
		}

		if (doafile(i, text, sendbuf, sblen, lang)==-1) {
			fprintf(stderr, "mkf: Unexpected EOF. Stop at %d/%d.\n", i, anum);
			exit(1);
		}
		free(text);
	}
	fclose(k);
	free(sendbuf);
	fclose(mkf_out);
	exit(0);
}

static char **words = NULL;
static int nwords;
static int wordssize = 0;

static void
add_word(w)
char *w;
{
	if (nwords>=wordssize) {
		wordssize += 8192;
		if (!(words = realloc(words, wordssize*sizeof *words))) {
			perror("mkf: realloc");
			exit(1);
		}
	}
	if (!(words[nwords++] = strdup(w))) {
		perror("mkf: strdup");
		exit(1);
	}
}

static void
free_words()
{
	int i;
	for (i=0; i<nwords; i++) {
		free(words[i]);
		words[i] = NULL;
	}
}

static int
compar(va, vb)
const void *va, *vb;
{
	const char * const *a, * const *b;
	a = va;
	b = vb;
	return strcmp(*a, *b);
}

static int
doafile(seq, text, sendbuf, sblen, lang)
char *text, *sendbuf;
char *lang;
{
	int i, j;

	nwords = 0;

	if (analyse_and_add(text, sendbuf, sblen, seq, lang)==-1) {
		fprintf(stderr, "mkf: anlyse error\n");
		return -1;
	}

	/* final stage: sort and write to file */
	if (mkf_type==RAWXR_TYPE_COUNT) {
		qsort(words, nwords, sizeof *words, compar);
		for (i=0; i<nwords; i = j) {
			for (j=i+1; j<nwords && !strcmp(words[i], words[j]); j++) ;
			fprintf(mkf_out, "%d %s\n", j-i, words[i]);
		}
		free_words();
	}
	fflush(mkf_out);
	/* we don't need to free words when mkf_type==(RAW || SIMULATE) */
	return 0;
}

static int
analyse_and_add(text, sendbuf, sblen, seq, lang)
char *text, *sendbuf;
char *lang;
{
	int ndllines;
	struct ps state;
	int loop;

	if ((ndllines = smap_count_ndllines(text))>sblen) {
		fprintf(stderr, "too many dllines\n");
		return -1;
	}
	for (loop=0, state.p=text, *state.escs = '\0'; state.p; loop++) {
		smap_get_phrase(&state, sendbuf, sblen-ndllines, 0);
		if (!sendbuf || !*sendbuf) {
			continue;
		}
		if (!*smap_skip_spaces_jis(sendbuf)) {
			continue;
		}
		if (analyse_a_phrase(lang, sendbuf)!=0) {
			fprintf(stderr, "analyse_a_phrase error\n");
			return -1;
		}
	}
if (loop>=2) fprintf(stderr, "get_phrase loop=%d\n", loop);
	return 0;
}

static int
analyse_a_phrase(lang, buf)
char *lang;
char *buf;
{
	int l;
	char wbuf[MAXWORDLEN*2+20];

	switch (mkf_type) {
	case RAWXR_TYPE_SIMULATE:
		fputs_sp(buf, mkf_out);
		break;
	case RAWXR_TYPE_COUNT:
	case RAWXR_TYPE_RAW:
		if (smap_set_lang(lang)==-1) {
			fprintf(stderr, "set_lang failed\n");
			return -1;
		}

		if (smap_start_text()==-1) {
			fprintf(stderr, "start_text failed\n");
			return -1;
		}

		smap_send_to_analyser(buf);
		for (l=1; smap_readln(wbuf, sizeof wbuf)>0; l++) {
			char *p;
			char *s1, *s2, *rest;
			p = wbuf;
			if (!strcmp(p, DOT)) {
				break;
			}
			if (*p=='.') p++;
			split(p, " ", &s1, &s2, &rest, (char **)NULL);
			if (!*s1 || !*s2) continue;
			if (mkf_type==RAWXR_TYPE_COUNT) {
				add_word(s2);
			}
			else {
				fprintf(mkf_out, " %s\n", s2);
			}
		}
		break;
	default:
		fprintf(stderr, "mkf: internal error 3\n");
		exit(1);
	}
	return 0;
}

static void
fputs_sp(s, stream)
char *s;
FILE *stream;
{
	int n;
	for (n=1; *s; s++) {
		if (n) {
			putc(' ', stream);
			n=0;
		}
		putc(*s, stream);
		if (*s=='\n') {
			n = 1;
		}
	}
	if (!n) {
		putc('\n', stream);
	}
}
