/*
 *	Copyright (C) 1998 Advanced Research Laboratory, Hitachi, Ltd.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: ana.c,v 1.9 2001/09/06 06:23:21 nis Exp $"; */

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/uio.h>
#include <sys/time.h>

#include <assert.h>
#include <errno.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <syslog.h>
#include <unistd.h>

#include <geta/iso2022jp2.h>
#include <geta/getamu.h>
#include <geta/srvmu.h>
#include <geta/smap.h>

#include "ssP.h"

static void writer(char *, int);
static void reader(char *, FILE *, int (*)(char *, size_t, char *), int);
static void sender(int);
static void addline(char *);
static char *fstrdup(char *);

static char *fullword = NULL;
static char **words = NULL;

static int nwords;
static int fullwordptr;

void
analyse(buf, s, c, bufsize)
char *buf;
struct _lang_conf *c;
{
	int status;
	pid_t inproc;
	FILE *f;
	int d;

	if (!fullword) {
		if (!(fullword = malloc(FULLWORDSIZE+32))) {
			syslog(LOG_ERR, "malloc fullword %s", strerror(errno));
			return;
		}
		if (!(words = malloc(MAXNWORDS * sizeof *words))) {
			syslog(LOG_ERR, "malloc words %s", strerror(errno));
			return;
		}
	}

	if (sem_lock(c->semidx)==-1) {
		fprintf(stderr, "c_text: sem_lock: lock failed\n");
		sem_status(c->semidx, stderr);
	}
	sstate(c->language);

/* note: if strlen(buf) is less than kern.maxsockbuf, fork is not essential.
   set SO_SNDBUF to the size and just write to the socket is OK. */
	switch (inproc = fork()) {
	case -1:
		syslog(LOG_ERR, "fork %s", strerror(errno));
		exit(1);
	case 0:
		writer(buf, c->fds[1]);
		exit(0);
	}

	nwords = 0;
	fullwordptr = 0;

	if ((d = dup(c->fds[0]))==-1) {
		syslog(LOG_ERR, "dup");
		exit(1);
	}
	if (!(f = fdopen(d, "r"))) {
		syslog(LOG_ERR, "fdopen");
		exit(1);
	}
	reader(buf, f, c->uniform, bufsize);
	fclose(f);

	sstate("main");

	if (sem_unlock(c->semidx)==-1) {
		fprintf(stderr, "c_text: sem_unlock: unlock failed\n");
		sem_status(c->semidx, stderr);
	}

	waitpid(inproc, &status, 0);
	sender(s);
}

extern int debug;

static void
writer(buf, d)
char *buf;
{
	char *p;
	struct iovec iov[2];

	for (p=buf; p=index(p, CSEP); *p++ = ' ') ;
	iov[0].iov_base = buf;
	iov[0].iov_len = strlen(buf);
	iov[1].iov_base = LF SEP LF;
	iov[1].iov_len = strlen(LF SEP LF);
	writev(d, iov, 2);
	if (debug & D_LOGINPUT) {
		FILE *fp;
		if (fp = fopen("/tmp/ss.input", "a")) {
			time_t t;
			t = time(NULL);
			fprintf(fp, "- %s", ctime(&t));
			fprintf(fp, "+ %s\n", buf);
			fclose(fp);
		}
	}
}

static void
reader(orig, f, uniform, bufsize)
char *orig;
FILE *f;
int (*uniform)(char *, size_t, char *);
{
	int ebs = bufsize/2-8;
	char *buf, *work;

	buf = malloc(bufsize+8);
	work = buf + ebs + 4;

	if (!buf) {
		perror("malloc");
		return;
	}

	while (buf[0] = '\0', fgets(buf, ebs, f), 1) {
		chop(buf);
		switch ((*uniform)(work, bufsize+8-ebs-4, buf)) {
			int l;
		case 0:
			/* normal word */
			l = strlen(work);
			if (l>=MAXWORDLEN*2+20) {
				/* ignore this word */
			}
			else {
				addline(work);
			}
			break;
		case 1:
			/* uniform found the delimiter */
			goto END;
		default:
			/* uniform cannot parse the line */
			/* do nothing */
			break;
		}
	}
  END: ;
	free(buf);
}

static void
sender(s)
{
	char *w, *e;
	int i;
	static char sendbuf[1024];
	char *p;

	e = sendbuf+sizeof sendbuf;
	for (i=0, p=sendbuf; i<nwords && (w = words[i]); i++) {
		int l;
		char *r;
		if (*w == '.') {
			r = "."; l = 1;
			if (p+l > e) {
				send(s, sendbuf, p-sendbuf, 0);
				p = sendbuf;
			}
			bcopy(r, p, l); p += l;
		}

		r = w, l = strlen(r);
		if (p+l > e) {
			send(s, sendbuf, p-sendbuf, 0);
			p = sendbuf;
		}
		if (l>sizeof sendbuf) {
			assert(p==sendbuf);
			send(s, r, l, 0);
		}
		else {
			bcopy(r, p, l); p += l;
		}

		r = CRLF; l = strlen(CRLF);
		if (p+l > e) {
			send(s, sendbuf, p-sendbuf, 0);
			p = sendbuf;
		}
		bcopy(r, p, l); p += l;
	}
	send(s, sendbuf, p-sendbuf, 0);
}

static char *
fstrdup(s)
char *s;
{
	int l;
	char *p;
	l = strlen(s)+1;
	if (fullwordptr + l > FULLWORDSIZE) {
		syslog(LOG_ERR, "fstrdup failed");
		return NULL;
	}
	p = fullword+fullwordptr;
	bcopy(s, p, l);
	fullwordptr += l;
	return p;
}

static void
addline(s)
char *s;
{
	if (nwords>=MAXNWORDS) {
		syslog(LOG_ERR, "addline: too many words.(%s)", s);
		return;
	}
	if (!(words[nwords++] = fstrdup(s))) {
		syslog(LOG_ERR, "addline: fullword is full.(%s)", s);
		return;
	}
}
