/*
 *	Copyright (C) 1998 Advanced Research Laboratory, Hitachi, Ltd.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: commands.c,v 1.10 2001/09/06 06:23:21 nis Exp $"; */

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/time.h>

#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <syslog.h>
#include <syslog.h>
#include <unistd.h>

#include <geta/getamu.h>
#include <geta/srvmu.h>
#include <geta/smap.h>
#include <geta/iso2022jp2.h>
#include <geta/jis.h>

#include "ssP.h"

static void c_helo(int, char **, int);
static void c_close(int, char **, int);
static void c_help(int, char **, int);
static void c_llang(int, char **, int);
static void c_lang(int, char **, int);
static void c_getbufsize(int, char **, int);
static void c_text(int, char **, int);
static void c_not_implemented(int, char **, int);

#define	OPENMSG		"200 SMAP %s (ss 1.0) Ready."
#define	HELOMSG		"200 ok"
#define	OKMSG		"200 ok"
#define	TEXTOKMSG	"320 Start text input"
#define	BUFSIZEMSG	"201 Buffer size follows"
#define	NOTAVAILMSG	"400 SMAP Service not available. [%s]"
#define	IFEWPARAMMSG	"401 too few arguments"
#define	IMANYPARAMMSG	"402 too many arguments"
#define	NOSUCHLANGMSG	"410 No such language"
#define	NOCURRLANGMSG	"411 No current language set"
#define	ERRMSG		"500 Syntax error, command unrecognized"
#define	NOTIMPLMSG	"502 command not implemented"

#define	sendmsg(s, msg)	send(s, msg, strlen(msg), 0)

/*
200 Server Ready
201 Buffer size follows
202 Help messages follows
203 Synchronization succeeded

210 Language has been set
211 Language list follows

320 Start text input

400 Service not available
403 Synchronization failed

410 No such language
411 No current language set

500 Syntax error, command unrecognized
501 Syntax error in parameters or arguments
502 Command not implemented
*/

struct command {
	char *name;
	void (*func)(int, char **, int);
	int minparam, maxparam;
	char *help;
};

struct command commands[] = {
	/* this array must be sorted */
	{"CLOSE", c_close, 0, 0, "close the session"},
	{"GETBUFSIZE", c_getbufsize, 0, 0, "get input buffer size"},
	{"HELO", c_helo, 0, 0, "synchoronize the transaction"},
	{"HELP", c_help, 0, 0, "show this message"},
	{"LANG", c_lang, 1, 1, "set current language"},
	{"LLANG", c_llang, 0, 0, "show list of available languages"},
	{"NOOP", c_helo, 0, 0, "echo status line"},
	{"TEXT", c_text, 0, 0, "start analyse"},
};

static struct _lang_conf *current_language = NULL;
static int ccompar(const void *va, const void *vb);

#define REALIBUFSIZE	IBUFSIZE+4

static char *ibuf;
static char *recvbuf;
static int rlen;

extern struct _lang_conf *lang_conf;
extern int nlanguages;

static char cs[513];

static int
ccompar(va, vb)
const void *va, *vb;
{
	const struct command *a, *b;
	a = va;
	b = vb;
	return strcasecmp(a->name, b->name);
}

void
do_server(s)
{
	char *rest;
	char hostname[MAXHOSTNAMELEN];

	gethostname(hostname, sizeof hostname);
	if (!(ibuf = malloc(REALIBUFSIZE))) {
		syslog(LOG_ERR, "do_server: return(malloc failed)");
		goto ERROR;
	}
	if (!(recvbuf = malloc(REALIBUFSIZE))) {
		syslog(LOG_ERR, "do_server: return(malloc failed)");
		goto ERROR;
	}
	snprintf(cs, sizeof cs, OPENMSG CRLF, hostname);
	sendmsg(s, cs);

	recvbuf[0] = '\0';

	while (recvlnb(s, recvbuf, 0, REALIBUFSIZE, CRLF, &rest, &rlen, 0)!=-1) {
		char *av[8];
		struct command key, *c;
		int ac;

		/* SMAP command line is not longer than 511(==sizeof cs -1) characters */
		scopy(recvbuf, cs, sizeof cs);
		bcopy(rest, recvbuf, rlen+1);

		ac = splitv(cs, " \t", av, nelems(av));
		key.name = av[0];
		
		if (c = bsearch(&key, commands, nelems(commands), sizeof *commands, ccompar)) {
			if (ac-1 < c->minparam) {
				sendmsg(s, IFEWPARAMMSG CRLF);
			}
			else if (c->minparam < ac-1) {
				sendmsg(s, IMANYPARAMMSG CRLF);
			}
			else {
				(*c->func)(ac, av, s);
			}
		}
		else {
			sendmsg(s, ERRMSG CRLF);
		}
	}
	free(ibuf);
	free(recvbuf);
	ibuf = NULL;
	recvbuf = NULL;
	return;

ERROR:
	snprintf(cs, sizeof cs, NOTAVAILMSG CRLF, hostname);
	free(ibuf);
	ibuf = NULL;
	free(recvbuf);
	recvbuf = NULL;
	sendmsg(s, cs);
	syslog(LOG_ERR, "do-server exited abnormaly");
	return;
}

static void
c_not_implemented(argc, argv, s)
char *argv[];
{
	sendmsg(s, NOTIMPLMSG CRLF);
}

static void
c_helo(argc, argv, s)
char *argv[];
{
	sendmsg(s, HELOMSG CRLF);
}

static void
c_close(argc, argv, s)
char *argv[];
{
	close(s);
}

static void
c_help(argc, argv, s)
char *argv[];
{
	struct command *c;
	for (c = commands; c<commands + nelems(commands); c++) {
		snprintf(cs, sizeof cs, "%s: %s" CRLF, c->name, c->help);
		sendmsg(s, cs);
	}
	sendmsg(s, DOT CRLF);
}

static void
c_llang(argc, argv, s)
char *argv[];
{
	int i;
	for (i=0; i<nlanguages; i++) {
		struct _lang_conf *c;
		c = &lang_conf[i];
		snprintf(cs, sizeof cs, "%s" CRLF, c->language);
		sendmsg(s, cs);
	}
}

static void
c_lang(argc, argv, s)
char *argv[];
{
	struct _lang_conf key, *c;
	key.language = argv[1];
	if (c = bsearch(&key, lang_conf, nlanguages, sizeof *lang_conf, lcompar)) {
		current_language = c;
		sendmsg(s, OKMSG CRLF);
	}
	else {
		sendmsg(s, NOSUCHLANGMSG CRLF);
	}
}

static void
c_getbufsize(argc, argv, s)
char *argv[];
{
	snprintf(cs, sizeof cs, BUFSIZEMSG CRLF "%d" CRLF, IBUFSIZE);
	sendmsg(s, cs);
}

static void
c_text(argc, argv, s)
char *argv[];
{
	char *p, *rest;

	if (!current_language) {
		sendmsg(s, NOCURRLANGMSG CRLF);
		return;
	}

	sendmsg(s, TEXTOKMSG CRLF);

	bcopy(recvbuf, ibuf, rlen+1);
	if (recvlnb(s, ibuf, rlen, REALIBUFSIZE, CRLF DOT CRLF, &rest, &rlen, 0)!=-1) {
		bcopy(rest, recvbuf, rlen+1);
		for (p=ibuf; p=strstr(p, CRLF DOT DOT); ) {
			p+=2;
			*p = ' ';
		}
		if (iso2022jp2p(ibuf)) {
			analyse(ibuf, s, current_language, REALIBUFSIZE);
		}
		else {
			fprintf(stderr, "invalid string\n");
		}
		sendmsg(s, DOT CRLF);
		return;
	}

	syslog(LOG_ERR, "too long input line. exit");
	exit(0);
}
