/*
 *	Copyright (C) 1998 Advanced Research Laboratory, Hitachi, Ltd.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: ifsmap.c,v 1.11 2001/09/06 06:23:21 nis Exp $"; */

#include <sys/types.h>
#include <sys/uio.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/time.h>
#include <netinet/in.h>

#include <ctype.h>
#include <netdb.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>

#include <geta/iso2022jp2.h>
#include <geta/jis.h>
#include <geta/gaiji.h>

#include <geta/getamu.h>
#include <geta/srvmu.h>
#include <geta/smap.h>
#include <geta/ifsmap.h>

static void smap_readln_init(void);
static int isabrk(int);
static int iskbrk(int, int, int, int);
#define	sendmsg(s, msg)	send(s, msg, strlen(msg), 0)

static int smap_socket = -1;

int
smap_setup(sblen, s)
int *sblen;
{
	char buf[513];
	char *s1, *rest;

	smap_socket = s;
	smap_readln_init();
	smap_readln(buf, sizeof buf);
	split(buf, " ", &s1, &rest, (char **)NULL);
	if (!s1 || atoi(s1)!=200) {
		fprintf(stderr, "cannot talk to smap server\n");
		return -1;
	}

	snprintf(buf, sizeof buf, "getbufsize" CRLF);
	sendmsg(smap_socket, buf);
	smap_readln(buf, sizeof buf);
	split(buf, " ", &s1, &rest, (char **)NULL);
	if (!s1 || atoi(s1)!=201) {
		fprintf(stderr, "cannot get buffer size\n");
		return -1;
	}
	smap_readln(buf, sizeof buf);
	*sblen = atoi(buf);
	return 0;
}

int
smap_set_lang(lang)
char *lang;
{
	char buf[513];
	char *s1, *rest;

	snprintf(buf, sizeof buf, "lang %s" CRLF, lang);
	sendmsg(smap_socket, buf);
	smap_readln(buf, sizeof buf);
	split(buf, " ", &s1, &rest, (char **)NULL);
	if (!s1 || atoi(s1)!=200) {
		fprintf(stderr, "cannot set language\n");
		return -1;
	}
	return 0;
}

int
smap_start_text()
{
	char buf[513];
	char *s1, *rest;

	scopy("text" CRLF, buf, sizeof buf);
	sendmsg(smap_socket, buf);
	smap_readln(buf, sizeof buf);
	split(buf, " ", &s1, &rest, (char **)NULL);
	if (!s1) {
		fprintf(stderr, "smap server is not ready -\n");
		return -1;
	}
	if (atoi(s1)!=320) {
		fprintf(stderr, "smap server is not ready \"%s\"\n", s1);
		return -1;
	}
	return 0;
}

void
smap_send_to_analyser(sendbuf)
char *sendbuf;
{
	char *p, *q;
	struct iovec iov[2];
	for (p=sendbuf; q=strstr(p, "\n."); p=q+2) {
		*(q+1) = '\0';
		iov[0].iov_base = p;
		iov[0].iov_len = strlen(p);
		iov[1].iov_base = "..";
		iov[1].iov_len = 2;
		writev(smap_socket, iov, 2);
	}
	iov[0].iov_base = p;
	iov[0].iov_len = strlen(p);
	iov[1].iov_base = CRLF DOT CRLF;
	iov[1].iov_len = strlen(CRLF DOT CRLF);
	writev(smap_socket, iov, 2);
}

/* dot de hazimaru gyou no kazu wo kazoeru */
int
smap_count_ndllines(p)
char *p;
{
	int ndllines;
	char *q;
	for (ndllines=0; q=strstr(p, "\n."); p=q+3, ndllines++) ;
	return ndllines;
}

void
smap_get_phrase(state, dst, len, rmjcrlf)
struct ps *state;
char *dst;
{
	int last = SHIFT_SB << 8 | 'B';
	char *dst0 = dst;
	int i;
	char *src, *end;
	int shift = SHIFT_SB;
	int shift_prev = 0;
	char f = 'B';
	struct ps bp[16];
	char escs[5];

#define	BP(I) (bp[I].p = src, scopy(escs, bp[I].escs, sizeof bp[I].escs), bp[I].d = dst)
	for (i=0; i<nelems(bp); i++) bp[i].d = NULL;
	end = dst+len-3;
	if (*state->escs) {
		char *e;
		shift_prev = shift;
		if ((shift=pescs(state->escs, &e))==SHIFT_ERR) goto ERR;
		f = *e;
		if (strcmp(state->escs, "\033(B")) {
			int el;
			el = e-state->escs+1;
			bcopy(state->escs, escs, el);
			bcopy(state->escs, dst, el);
			dst += el;
			escs[el] = 0;
		}
	}
	else {
		/* use default value of f and shift */
		scopy("\033(B", escs, sizeof escs);
	}
	for (src=state->p; *src && dst<end; ) {
		if (*src==033) {
			char *e;
			int el;
			shift_prev = shift;
			if ((shift=pescs(src, &e))==SHIFT_ERR) goto ERR;
			f = *e;
			el = e-src+1;
			bcopy(src, escs, el);
			bcopy(src, dst, el);
			dst += el;
			escs[el] = 0;
			src = e+1;
		}
		else {
			switch (shift) {
			case SHIFT_MB:
				if (f=='@' || f=='B') {
					int h, l;
					int n1, n2;
					int bl;
					last = SHIFT_MB << 8 | 'B';
					h = *src++;
					l = *src++;
					if (!l) goto ERR;
					*dst++ = h;
					*dst++ = l;
					n1 = *src;
					n2 = *(src+1);
					if ((bl = iskbrk(h, l, n1, n2))>=0) {
						BP(bl);
					}
				}
				else {
					last = SHIFT_MB << 8 | f;
					if (!*(src+1)) goto ERR;
					*dst++ = *src++;
					*dst++ = *src++;
				}
				break;
			case SHIFT_SB:
				if (f=='B' || f=='J') {
					int c;
					int bl;
					c = *src;

					if (rmjcrlf && (last == (SHIFT_MB<<8|'B') && (c=='\r' || c=='\n'/* || c==' '*/))) {
						src++;
					}
					else {
						last = SHIFT_SB << 8 | f;
						*dst++ = *src++;
						if (c=='\n') BP(1);
						if (isspace(c)) BP(2);
						if ((bl = isabrk(c))>=0) BP(bl);
					}
				}
				else {
					last = SHIFT_SB << 8 | f;
					*dst++ = *src++;
				}
				break;
			case SHIFT_SS:
				shift = shift_prev;
				last = SHIFT_SS << 8 | f;
				*dst++ = *src++;
				break;
			default:
				break;
			}
		}
		if (*src==033) BP(6);
		else BP(15);
	}
ERR:
	if (dst<end) {
		src = NULL;
		BP(0);
	}
	for (i=0; i<nelems(bp); i++) {
		if (bp[i].d) {
			char *d;
			d = bp[i].d;
if (i>=2) fprintf(stderr, "break-level = %d\n", i);
			*d = '\0';
			if (strcmp(bp[i].escs, "\033(B")) {
				scopy("\033(B", d, 4);	/* ? */
			}
			state->p = bp[i].p;
			scopy(bp[i].escs, state->escs, sizeof state->escs);
			if (dst>=dst0+len || dst<=dst0) {
				state->p = NULL;
				*dst = '\0';
				return;
			}
			return;
		}
	}
	state->p = NULL;
	*dst = '\0';
	return;
}

static int
isabrk(c)
{
	if (c==',') return 3;
	if (index("()[]{}", c)) return 7;
	if (index("!\"#$%&*+-/:;<=>?@\\^_|~", c)) return 10;
	if (index("'", c)) return 13;
	return -1;
}

static int
iskbrk(h, l, h2, l2)
{
	h-=0x20;
	l-=0x20;
	h2-=0x20;
	l2-=0x20;

#if defined HDH_CUSTOM
	/* HDH Gaiji Indicator: cannot break */
	if (h2==GHKU && l2==GHTEN) return -1;
#endif

	/* touten, kuten */
	if ((h==1 && l==2) || (h==1 && l==3)) return 4;

	if (h2==033) return 5;

	/* hiragana, katakana, greek, kiriru + kanji */
	if ((h==4 || h==5 || h==6 || h==7) && (16 <= h2)) return 14;

	return -1;
}

char *
smap_skip_spaces_jis(s)
char *s;
{
	int shift;
	char f;

	shift = SHIFT_SB;
	f = 'B';
	for (; *s; ) {
		if (*s==033) {
			int ns;
			char nf;
			char *e;
			if ((ns=pescs(s, &e))==SHIFT_ERR) goto ERR;
			nf = *e;
			s = e+1;

			if (ns==SHIFT_SS) {
				goto END;
			}
			else {
				shift = ns;
				f = nf;
			}
		}
		else {
			switch (shift) {
			case SHIFT_MB:
				goto END;
			case SHIFT_SB:
				if (f!='B') goto END;
				if (!isspace(*s)) goto END;
				s++;
				break;
			}
		}
	}
  ERR: ;
  END: ;
	return s;
}

static char rbuf[8192];
static char *rbp = NULL;
static char *rbe = NULL;

static void
smap_readln_init()
{
	rbp = rbe = NULL;
}

static int
filbuf()
{
	int c;
	if ((c = recv(smap_socket, rbuf, sizeof rbuf, 0))==-1) {
		return -1;
	}
	rbp = rbuf;
	rbe = rbuf + c;
	return c;
}

ssize_t
smap_readln(buf, n)
char *buf;
size_t n;
{
	char *bp, *be;

	if (!rbp || rbp>=rbe) {
		filbuf();
	}

	if (!n) {
		return -1;
	}

	for (bp=buf, be=buf+n-1; bp<be && rbp<rbe; ) {
		if ((*bp++ = *rbp++)=='\n') {
			break;
		}
		if (rbp>=rbe) {
			filbuf();
		}
	}
	*bp = '\0';
	if (bp>buf && *(bp-1)=='\n') {
		*--bp = '\0';
		if (bp>buf && *(bp-1)=='\r') {
			*--bp = '\0';
		}
	}
	return bp-buf;
}
