/*
 *	Copyright (C) 1998 Advanced Research Laboratory, Hitachi, Ltd.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: sc.c,v 1.9 2001/09/06 06:23:21 nis Exp $"; */

/* A simple smap client program. */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>

#include <netdb.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>

#include <geta/getamu.h>
#include <geta/srvmu.h>
#include <geta/smap.h>
#include <geta/ifsmap.h>

static void do_morph(char *, char *, int, char *);
static void analyse_a_phrase(char *, char *);

char *progname;

static void usage(void);

static void
usage()
{
	fprintf(stderr, "usage: %s (-l local_socket|-i server:port) lang\n", progname);
	exit(1);
}

main(argc, argv)
char *argv[];
{
	int s;
	char *localsocket = NULL;
	int type = AF_UNSPEC;
	char *sendbuf;
	char *text;
	char *server_host = NULL;
	int sblen;
	int port = 0;
	char *lang;

	int host;
	struct hostent *h;

	if (progname=rindex(argv[0], '/')) {
		progname++;
	}
	else {
		progname = argv[0];
	}

	if (argc!=4) {
		usage();
	}

	if (!strcmp(argv[1], "-l")) {
		localsocket = argv[2];
		type = AF_UNIX;
	}
	else if (!strcmp(argv[1], "-i")) {
		char *p;
		server_host = argv[2];
		if (!(p = rindex(server_host, ':'))) {
			usage();
		}
		*p = '\0';
		port = atoi(p+1);
		type = AF_INET;
	}
	else {
		usage();
	}
	lang = argv[3];

	if (type==AF_UNIX) {
		if ((s = un_connect_to_server(localsocket))==-1) {
			fprintf(stderr, "cannot connect to server\n");
			return 1;
		}
	}
	else if (type==AF_INET) {
		if (!(h = gethostbyname(server_host))) {
			perror("gethostbyname");
			return 1;
		}
		host = *(int *) *h->h_addr_list;
		if ((s = in_connect_to_server(host, port))==-1) {
			fprintf(stderr, "cannot connect to server\n");
			return 1;
		}
	}
	else {
		fprintf(stderr, "sc: internal error\n");
		return 1;
	}

	if (smap_setup(&sblen, s)==-1) {
		fprintf(stderr, "smap_setup failed\n");
		return 1;
	}

	if (!(text = readin(stdin))) {
		fprintf(stderr, "readin failed\n");
		return 1;
	}
	if (!(sendbuf = malloc(sblen))) {
		perror("malloc");
		return 1;
	}
	do_morph(text, sendbuf, sblen, lang);
	free(sendbuf);
	free(text);
	return 0;
}

static void
do_morph(text, sendbuf, sblen, lang)
char *text;
char *sendbuf;
char *lang;
{
	struct ps state;
	int ndllines;

	if ((ndllines = smap_count_ndllines(text))>sblen) {
		fprintf(stderr, "too many dllines\n");
		return;
	}

	for (state.p=text, *state.escs = '\0'; state.p; ) {
		smap_get_phrase(&state, sendbuf, sblen-ndllines, 0);
		if (!sendbuf || !*sendbuf) {
			continue;
		}
		if (!*smap_skip_spaces_jis(sendbuf)) {
			continue;
		}
		analyse_a_phrase(lang, sendbuf);
	}
}

static void
analyse_a_phrase(lang, buf)
char *lang, *buf;
{
	char wbuf[MAXWORDLEN*2+20];
	if (smap_set_lang(lang)==-1) {
		fprintf(stderr, "set_lang failed\n");
		return;
	}
	if (smap_start_text()==-1) {
		fprintf(stderr, "start_text failed\n");
		return;
	}
	smap_send_to_analyser(buf);
	for (; smap_readln(wbuf, sizeof wbuf)>0; ) {
		char *p;
		p = wbuf;
		if (!strcmp(p, DOT)) {
			break;
		}
		if (*p=='.') p++;
		printf("%s\n", p);
	}
}
