/*
 *	Copyright (C) 1999 Advanced Research Laboratory, Hitachi, Ltd.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: ssconfig.c,v 1.8 2001/09/06 06:23:21 nis Exp $"; */

#include <sys/param.h>
#include <sys/stat.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <geta/getamu.h>
#include <geta/srvmu.h>
#include <geta/smap.h>
#include "ssP.h"

struct builtin_analizer {
	char *name;
	void (*analyse)(void);
};

struct uniform_filter {
	char *name;
	int (*uniform)(char *, size_t, char *);
};

#define	NPAIR(x)	{#x, x}

/* builtin_analizers, uniform_filters ha initialization no toki ni dake
   linear search sarerunode, aete bsearch niha sinai. */

static struct builtin_analizer builtin_analizers[] = {
	NPAIR(do_split),
	{NULL, NULL}
};

static struct uniform_filter uniform_filters[] = {
	NPAIR(split1),
	NPAIR(splitn),
	NPAIR(stemn),
	NPAIR(ngramn),
	NPAIR(chasenn),
	NPAIR(animar),
	NPAIR(def_filt),
	{NULL, NULL}
};

struct _lang_conf *lang_conf;
int nlanguages = 0;

int
readconf(configfile, projectroot)
char *configfile, *projectroot;
{
	FILE *f;
	int i, l;
	struct _lang_conf *p;
	char buf[MAXPATHLEN+128];

	if (!(f = fopen(configfile, "r"))) {
		perror(configfile);
		return -1;
	}
	for (i=0; fgets(buf, sizeof buf, f); ) {
		chop(buf);
		if (isspaces(buf) || *buf=='#') {
			continue;
		}
		i++;
	}
	rewind(f);
	lang_conf = calloc(i+1, sizeof *lang_conf);
	
	for (l=1, p=lang_conf; fgets(buf, sizeof buf, f); l++) {
		char *name, *file_name, *analyser_name, *uniform_name;
		struct builtin_analizer *bp;
		struct uniform_filter *up;

		chop(buf);
		if (isspaces(buf) || *buf=='#') {
			continue;
		}
		split(buf, ":", &name, &file_name, &uniform_name, (char **)NULL);

		if (*file_name == '*') {
			analyser_name = file_name+1;
			file_name = NULL;
		}
		else {
			analyser_name = NULL;
		}

		p->language = strdup(name);
		if (file_name) {
			struct stat sb;
			char path[MAXPATHLEN];
			if (*file_name=='/') {
				scopy(file_name, path, sizeof path);
			}
			else {
				snprintf(path, sizeof path, "%s/%s", projectroot, file_name);
			}
			if (stat(path, &sb)==-1) {
				fprintf(stderr, "malformed %s. line=%d, %s: no such file.\n", configfile, l, path);
				return -1;
			}
			p->start = strdup(path);
			p->analyse = NULL;
		}
		else {
			for (bp=builtin_analizers; bp->name; bp++) {
				if (!strcmp(bp->name, analyser_name)) {
					break;
				}
			}
			if (!bp->name) {
				fprintf(stderr, "malformed %s. line=%d, %s: no such built in function.\n", configfile, l, analyser_name);
				return -1;
			}
			p->start = NULL;
			p->analyse = bp->analyse;
		}
		for (up=uniform_filters; up->name; up++) {
			if (!strcmp(up->name, uniform_name)) {
				break;
			}
		}
		if (!up->name) {
			fprintf(stderr, "malformed %s. line=%d, %s: no such built in filter.\n", configfile, l, uniform_name);
			return -1;
		}
		p->uniform = up->uniform;
		p++;
	}

	fclose(f);
	nlanguages = p-lang_conf;
	qsort(lang_conf, nlanguages, sizeof *lang_conf, lcompar);
	return 0;
}

int
lcompar(va, vb)
const void *va, *vb;
{
	const struct _lang_conf *a, *b;
	a = va;
	b = vb;
	return strcasecmp(a->language, b->language);
}
